//
//  SWMultiColumnTableView.m
//  SWGameLib
//
//
//  Copyright (c) 2010 Sangwoo Im
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
//  
//  Created by Sangwoo Im on 6/3/10.
//  Copyright 2010 Sangwoo Im. All rights reserved.
//

#import "SWMultiColumnTableView.h"
#import "SWTableViewCell.h"
#import "CGPointExtension.h"
#import "SWDebug.h"

@interface SWTableView ()

-(NSUInteger)_indexFromOffset:(CGPoint)offset;
-(CGPoint)_offsetFromIndex:(NSUInteger)index;
-(void)_updateContentSize;

@end

@implementation SWMultiColumnTableView
@synthesize colCount;

-(id)initWithViewSize:(CGSize)size {
    if ((self = [super initWithViewSize:size])) {
        colCount = 1;
    }
    return self;
}
-(void)setColCount:(NSUInteger)cols {
    colCount = cols;
    if (self.direction == SWScrollViewDirectionBoth) {
        [self _updateContentSize];
    } else {
        SWLog(@"%s: %s: You cannot set this value. The value will ignored and overwritten.", __FILE__, __FUNCTION__);
    }
}
-(NSUInteger)_indexFromOffset:(CGPoint)offset {
    NSInteger  index;
    CGSize     cellSize;
    NSUInteger col, row;
    CGFloat    spaceWidth;
    
    cellSize = [self.dataSource cellSizeForTable:self];
    
    switch (self.direction) {
        case SWScrollViewDirectionBoth:
            spaceWidth = self.contentSize.width / colCount;
            col        = (offset.x - (spaceWidth - cellSize.width)*0.5)/spaceWidth;
            row        = offset.y / cellSize.height;
            break;
        case SWScrollViewDirectionHorizontal:
            colCount   = self.viewSize.height / cellSize.height;
            spaceWidth = self.viewSize.height / colCount;
            col        = (offset.y - (spaceWidth - cellSize.height)*0.5)/spaceWidth;
            row        = offset.x / cellSize.width;
            break;
        default:
            colCount   = self.viewSize.width / cellSize.width;
            spaceWidth = self.viewSize.width / colCount;
            col        = (offset.x - (spaceWidth - cellSize.width)*0.5)/spaceWidth;
            row        = offset.y / cellSize.height;
            break;
    }
    index = col + row * colCount;
    index = MAX(0, index);
    index = MIN(((NSInteger)[self.dataSource numberOfCellsInTableView:self])-1, index);
    return ((NSUInteger)index);
}
-(CGPoint)_offsetFromIndex:(NSUInteger)index {
    CGPoint    offset;
    CGSize     cellSize;
    CGFloat    spaceWidth;
    NSUInteger col, row;
    
    SWAssert(index != NSNotFound, @"CCTableView: _offsetFromIndex: invalid index");
    
    cellSize = [self.dataSource cellSizeForTable:self];
    switch (self.direction) {
        case SWScrollViewDirectionBoth:
            row        = index / colCount;
            col        = index % colCount;
            spaceWidth = self.contentSize.width / colCount;
            offset     = ccp(col * spaceWidth + (spaceWidth - cellSize.width) * 0.5,
                             row * cellSize.height);
            break;
        case SWScrollViewDirectionHorizontal:
            colCount   = self.viewSize.height / cellSize.height;
            row        = index / colCount;
            col        = index % colCount;
            spaceWidth = self.viewSize.height / colCount;
            offset     = ccp(row * cellSize.height,
                             col * spaceWidth + (spaceWidth - cellSize.width) * 0.5);
            break;
        default:
            colCount   = self.viewSize.width / cellSize.width;
            row        = index / colCount;
            col        = index % colCount;
            spaceWidth = self.viewSize.width / colCount;
            offset     = ccp(col * spaceWidth + (spaceWidth - cellSize.width) * 0.5,
                             row * cellSize.height);
            break;
    }
    
    return offset;
}
-(void)_updateContentSize {
    CGSize     size, cellSize;
    NSUInteger cellCount, rows;
    
    cellSize  = [self.dataSource cellSizeForTable:self];
    cellCount = [self.dataSource numberOfCellsInTableView:self];
    switch (self.direction) {
        case SWScrollViewDirectionBoth:
            rows     = ceilf(cellCount/((CGFloat)colCount));
            size     = CGSizeMake(cellSize.width * colCount, rows * cellSize.height);
            break;
        case SWScrollViewDirectionHorizontal:
            colCount = self.viewSize.height / cellSize.height;
            rows     = ceilf(cellCount/((CGFloat)colCount));
            size     = CGSizeMake(rows * cellSize.width, colCount * cellSize.height);
            break;
        default:
            colCount = self.viewSize.width / cellSize.width;
            rows     = ceilf(cellCount/((CGFloat)colCount));
            size     = CGSizeMake(cellSize.width * colCount, rows * cellSize.height);
            break;
    }
    [self setContentSize:size];
}
@end
